import React from "react";
import './highlight.css'
import {compose} from "./utils";
// import PropTypes from 'prop-types'

const getMarkTagLength=()=>"<mark></mark>".length
const markTagOffset = getMarkTagLength();
const tag = 'MARK'


export default class Highlight extends React.Component {
    constructor(props) {
        super(props);
        this.historyRanges = [];
        this.state={}
        this.extractAndSetInitialValues = compose(this.setInitialValues, this.extractDetailsFrom)
    }

    onMouseUpHandler=(index)=>{
        const selectionObj = (window.getSelection && window.getSelection());
        if (selectionObj.isCollapsed) return;
        const {parentNode, focusParentNode, anchorNode,
            focusOffset, anchorOffset, focusNode, selection , forward} = this.extractAndSetInitialValues(selectionObj);
        if (parentNode !== focusParentNode) return;
        let {endOffset, startOffset} = this.extractStartEndOffSets(parentNode, focusNode, anchorNode, focusOffset, anchorOffset);
        this.callSelectionHandler(startOffset, endOffset, forward, selection, index);
    }

    extractStartEndOffSets(parentNode, focusNode, anchorNode, focusOffset, anchorOffset) {
        let transformedNodes = [...parentNode.childNodes].map((node, i) => this.substituteNode(parentNode, i));
        let [focusNodeIndex, anchorNodeIndex] = [
            transformedNodes.findIndex(n => n === focusNode),
            transformedNodes.findIndex(n => n === anchorNode)
        ];
        let endOffset = transformedNodes.slice(0, focusNodeIndex + 1).reduce((acc, node) => this.updateEndOffset(node, focusNode, acc, focusOffset), 0);
        let startOffset = transformedNodes.slice(0, anchorNodeIndex + 1).reduce((acc, node) => this.updateStartOffset(node, anchorNode, acc, anchorOffset), 0);
        return {endOffset, startOffset};
    }

    selectionHandler = (selectionStart, selectionEnd, paragraphIndex) => {
        let paragraph = this.state.paragraphs ? [...this.state.paragraphs] : [...this.props.paragraphs];
        let paragraphString = this.props.paragraphs[paragraphIndex];
        this.mergeRanges({start: selectionStart, end: selectionEnd}, paragraphIndex);
        paragraphString = this.updateParagraphString(paragraphIndex, paragraphString);
        paragraph[paragraphIndex] = paragraphString;
        this.setState({paragraphs: paragraph})
    }


    mergeRanges = (range, paraIndex) => {
        if (this.historyRanges[paraIndex] == null) this.historyRanges[paraIndex] = [];
        let mid = Math.floor((this.historyRanges[paraIndex].length) / 2);
        let index = this.historyRanges[paraIndex].length > 0 ? this.binarySearch(0, this.historyRanges[paraIndex].length - 1, mid, range, paraIndex) : mid;
        let merged = this.finallyMerge(paraIndex, range);
        if (!merged) this.historyRanges[paraIndex] = [...this.historyRanges[paraIndex].slice(0, index),
            range,
            ...this.historyRanges[paraIndex].slice(index, this.historyRanges[paraIndex].length)
        ];
    }

    mergeifMergeable = (source, target) => {
        let howToMerge = (source['start'] <= target['start'] && target['start'] <= source['end'] ? 1 :
            target['end'] <= source['end'] && target['end'] >= source['start'] ? 2 :
                target['start'] <= source['start'] && target['end'] >= source['end'] ? 3 : 0);
        if (howToMerge === 1) {
            return {start: source['start'], end: Math.max(target['end'], source['end'])}
        } else if (howToMerge === 2) {
            return {start: Math.min(source['start'], target['start']), end: source['end']}
        } else if (howToMerge === 3) {
            return target;
        }
        return false;
    }

    binarySearch = (start, end, mid, newEle, paraIndex) => {

        while (start < end) {
            let ele = this.historyRanges[paraIndex][mid];
            if (ele['start'] === newEle['start']) {
                return mid;
            } else if (ele['start'] > newEle['start']) {
                end = mid - 1
                if (end < start) break;
                mid = Math.ceil((start + end) / 2);
            } else {
                start = mid + 1;
                if (start > end) break;
                mid = Math.ceil((start + end) / 2);
            }
        }
        if (newEle['start'] > this.historyRanges[paraIndex][mid]['start']) {
            return mid + 1;
        } else {
            return mid;
        }
    }

    finallyMerge(paraIndex, range) {
        let merged = false;
        this.historyRanges[paraIndex].forEach((src, i)=>{
            let target = merged ? this.historyRanges[paraIndex][i - 1] : range;
            let merge = this.mergeifMergeable(src, target);
            if (merge) {
                this.historyRanges[paraIndex][i] = merge;
                merged = true;
            }
        })
        return merged;
    }

    updateParagraphString(paragraphIndex, paragraphString) {
        this.historyRanges[paragraphIndex].forEach((range, index) => {
            let start = range['start'] + (index * markTagOffset);
            let end = range['end'] + (index * markTagOffset);
            paragraphString = `${paragraphString.slice(0, start)}<mark>${paragraphString.slice(start, end + 1)}</mark>${paragraphString.slice(end + 1, paragraphString.length)}`;
        });
        return paragraphString;
    }

    callSelectionHandler(startOffset, endOffset, forward, selection, index) {
        let {selectionStart, selectionEnd}  = this.updateSelectionEndpoints(forward, selectionStart, endOffset, selectionEnd, startOffset);
        this.selectionHandler(selectionStart, selectionEnd - 1, index);
    }


    updateSelectionEndpoints(forward, selectionStart, endOffset, selectionEnd, startOffset) {
        return !forward ? {selectionStart:endOffset, selectionEnd:startOffset} :  {selectionStart:startOffset, selectionEnd:endOffset}
    }

    updateStartOffset(node, anchorNode, startOffset, anchorOffset) {
        if (node !== anchorNode) {
            startOffset += node.nodeType === 1 ? node.innerText.length : node.length;
        }
        if (node === anchorNode) {
            startOffset += anchorOffset;
        }
        return startOffset;
    }

    updateEndOffset(node, focusNode, endOffset, focusOffset) {
        if (node !== focusNode) {
            endOffset += node.nodeType === 1 ? node.innerText.length : node.length;
        }
        if (node === focusNode) {
            endOffset += focusOffset;
        }
        return endOffset ;
    }

    substituteNode(parentNode, i) {
        let node = parentNode.childNodes[i];
        if (node.nodeName === tag) {
            node = node.childNodes[0];
        }
        return node;
    }

    getParentAndAnchorNode(anchorNode, focusNode) {
        let parentNode = anchorNode.parentNode;
        let focusParentNode = focusNode.parentNode;
        //
        if (anchorNode.parentElement.nodeName === tag) {
            parentNode = parentNode.parentNode;
        }
        if (focusNode.parentElement.nodeName === tag) {
            focusParentNode = focusParentNode.parentNode;
        }
        return { parentNode, focusParentNode };
    }

    setInitialValues=({anchorNode, focusNode, ...rest})=> {
        let { parentNode, focusParentNode } = this.getParentAndAnchorNode(anchorNode, focusNode);
        return { parentNode, focusParentNode, ...rest, anchorNode, focusNode };
    }

    isForward(position, anchorNode, focusOffset, anchorOffset) {
        //checks if the selection is ltr or rtl
        let forward = false;
        if (position === anchorNode.DOCUMENT_POSITION_FOLLOWING) {
            forward = true;
        } else if (position === 0) {
            forward = (focusOffset - anchorOffset) > 0;
        }
        return forward;
    }

    extractDetailsFrom=(selectionObj)=> {
        const selection = selectionObj.toString();
        const {anchorOffset, focusOffset, anchorNode, focusNode} = selectionObj;
        const position = anchorNode.compareDocumentPosition(focusNode);
        let forward = this.isForward(position, anchorNode, focusOffset, anchorOffset);
        return { position, anchorNode, focusOffset, anchorOffset, focusNode, selection, forward };
    }

    render() {
        let p = this.state.paragraphs || this.props.paragraphs;
        return(
            <>
                {
                    p.map((paragraph,i) => {
                        return <p dangerouslySetInnerHTML={{__html: paragraph}}
                                  key={i} className={`${this.props.className || ""}`} onMouseUp={(e) => {
                            e.preventDefault();
                            this.onMouseUpHandler(i)
                        }}/>
                    })
                }
            </>
        )
    }

}

// Highlight.defaultProps = {
//     className:"",
//     paragraphs:[]
// }
//
// Highlight.propTypes = {
//     paragraphs: PropTypes.arrayOf(string),
//     className:PropTypes.string
// }

